from chainer.training import Extension
from chainer.training import trigger as trigger_module


class SwitchTrainables(Extension):

	def __init__(self, switch_epochs, model, pooling):
		self.switch_epochs = switch_epochs
		self._trigger = trigger_module.get_trigger((switch_epochs, "epoch"))

		self.model = model
		self.pooling = pooling

	def __call__(self, trainer):
		if not self._trigger(trainer): return

		epoch = trainer.updater.epoch

		if (epoch / self.switch_epochs) % 2 == 0:
			self.model.enable_update()
			self.pooling.disable_update()
		else:
			self.model.disable_update()
			self.pooling.enable_update()

	def initialize(self, trainer):
		self.pooling.disable_update()




import numpy as np

from .base import GPUMixin

from sklearn.mixture.gaussian_mixture import GaussianMixture

class GMM(GPUMixin, GaussianMixture):


	def _m_step(self, X, log_resp, xp=np):
		""" M-Step
			Copied from sklearn/mixture/gaussian_mixture.py
		"""

		self.weights_, self.means_, self.covariances_ = \
			self._gaussian_params(X, log_resp, xp=xp)

		self.weights_ /= X.shape[0]



from .observable_linear import *
from .pooling import *

import chainer
import chainer.functions as F
import chainer.links as L

class Conv2D_BN(chainer.Chain):
	def __init__(self, insize, outsize, ksize,
		stride=1, pad=0,
		activation=F.relu,
		nobias=True,
		use_gamma=False,
		eps=2e-5):

		super(Conv2D_BN, self).__init__()
		assert callable(activation)

		with self.init_scope():
			self.conv = L.Convolution2D(insize, outsize,
				ksize=ksize, stride=stride, pad=pad, nobias=nobias)
			self.bn = L.BatchNormalization(outsize,
				use_gamma=use_gamma, eps=eps)

		self.activation = activation

	def forward(self, x):
		x = self.conv(x)
		x = self.bn(x)
		return self.activation(x)



#!/usr/bin/env python
if __name__ != '__main__': raise Exception("Do not import me!")
import sys

sys.path.append("..")
from inception import InceptionV3

import numpy as np
import chainer

from chainer.serializers import save_npz

pretrained_model = "/home/korsch/Data/MODELS/inception/ft_inat/model.ckpt.npz"
if len(sys.argv) >= 2:
	pretrained_model = sys.argv[1]

model = InceptionV3(pretrained_model, n_classes=1000)

save_npz("/home/korsch/Data/MODELS/inception/ft_inat/model.npz", model)

model.to_gpu(0)
var = model.xp.array(np.random.randn(8, 3, 299, 299).astype(np.float32))

with chainer.using_config("train", False):
	pred = model(var)

import pdb; pdb.set_trace()



import numpy as np

from functools import reduce, partial, wraps
from collections.abc import Iterable

from chainercv.transforms import center_crop, resize, scale

class Size(object):
	dtype=np.int32

	def __init__(self, value):
		self._size = np.zeros(2, dtype=self.dtype)
		if isinstance(value, int):
			self._size[:] = value

		elif isinstance(value, Size):
			self._size[:] = value._size

		elif isinstance(value, Iterable):
			assert len(value) <= 2, \
				"only iterables of maximum size 2 are supported, but was {}!".format(len(value))
			self._size[:] = np.round(value)


		else:
			raise ValueError("Unsupported data type: {}!".format(type(value)))

	def __str__(self):
		return "<Size {}x{}>".format(*self._size)

	def __repr__(self):
		return str(self)

	def __add__(self, other):
		return Size(self._size + other)

	def __sub__(self, other):
		return Size(self._size - other)

	def __mul__(self, other):
		return Size(self._size * other)

	def __truediv__(self, other):
		return Size(self._size / other)

	def __floordiv__(self, other):
		return Size(self._size // other)

	def __iter__(self):
		return iter(self._size)

	def __len__(self):
		return len(self._size)

def asarray(func):

	@wraps(func)
	def inner(image, *args, **kw):
		from PIL import Image
		is_pil = isinstance(image, Image.Image)
		if is_pil:
			image = np.asarray(image)
		image = func(image, *args, **kw)
		if is_pil:
			image = Image.fromarray(image)
		return image

	return inner

def crop(image, x,y,w,h):
	return image[:, y: y + h, x: x + w]

@asarray
def random_horizontal_flip(image, threshold=.5, axis=2):
	if np.random.random() < threshold:
		return np.flip(image, axis=axis)
	return image

@asarray
def random_crop(image, crop_size):
	th, tw = crop_size
	c, h, w = image.shape
	th, tw = min(th, h), min(tw, w)
	x = np.random.randint(0, w - tw)
	y = np.random.randint(0, h - th)
	return crop(image, x, y, tw, th)

@asarray
def _center_crop(image, crop_size):
	th, tw = crop_size
	c, h, w = image.shape
	th, tw = min(th, h), min(tw, w)
	x = int((w - tw) / 2.)
	y = int((h - th) / 2.)
	return crop(image, x, y, tw, th)


def generic_prepare(im, size,
		crop_fraction=0.875,
		swap_channels=True,
		zero_mean=False,
		keep_ratio=True
	):

	crop_size = None
	h, w, c = im.shape


	_im = im.transpose(2, 0, 1)

	if swap_channels:
		# RGB -> BGR
		_im = _im[::-1]

	if crop_fraction:
		crop_size = (np.array([h, w]) * crop_fraction).astype(np.int32)
		_im = center_crop(_im, crop_size)

	# bilinear interpolation
	if keep_ratio:
		if isinstance(size, tuple):
			size = size[0]
		_im = scale(_im, size, interpolation=2)
	else:
		if isinstance(size, int):
			size = (size, size)
		_im = resize(_im, size, interpolation=2)

	if _im.dtype == np.uint8:
		# rescale [0 .. 255] -> [0 .. 1]
		_im = (_im / 255).astype(np.float32)


	if zero_mean:
		# rescale [0 .. 1] -> [-1 .. 1]
		_im = _im * 2 - 1

	return _im


def generic_tf_prepare(size, crop_fraction=0.875, from_path=False):

	import tensorflow as tf
	config_sess = tf.ConfigProto(allow_soft_placement=True)
	config_sess.gpu_options.allow_growth = True
	sess = tf.Session(config=config_sess)

	if from_path:
		im_input = tf.placeholder(tf.string)
		image = tf.image.decode_jpeg(tf.read_file(im_input), channels=3)
		image = tf.image.convert_image_dtype(image, tf.float32)
	else:
		image = im_input = tf.placeholder(tf.float32, shape=(None, None, 3))


	raise NotImplementedError("REFACTOR ME!")
	image = tf.image.central_crop(image, central_fraction=crop_fraction)
	image = tf.expand_dims(image, 0)
	image = tf.image.resize_bilinear(image, [size, size], align_corners=False)
	image = tf.squeeze(image, [0])
	image = tf.subtract(image, 0.5)
	output = tf.multiply(image, 2)

	def inner(im):
		if not from_path and im.dtype == np.uint8:
			im = im / 255

		res = sess.run(output, feed_dict={im_input: im})
		return res.transpose(2, 0, 1)

	return inner


class Augmentation(object):
	def __init__(self):
		self.augmentations = []

	def __call__(self, image):
		if not self.augmentations: return image
		def call(val, func):
			return func(val)
		return reduce(
			call, self.augmentations,
			image)

	def center_crop(self, size):
		self.augmentations.append(
			partial(_center_crop, crop_size=size))
		return self

	def random_crop(self, size):
		self.augmentations.append(
			partial(random_crop, crop_size=size))
		return self

	def random_horizontal_flip(self, threshold=.5, axis=2):

		self.augmentations.append(
			partial(random_horizontal_flip, threshold=threshold, axis=axis))
		return self



from chainer_addons.training.optimizer import OptimizerType, optimizer, lr_shift, no_grad
from chainer_addons.training.extensions.progress_bar import JupyterProgressBar
from chainer_addons.training.updater import MiniBatchUpdater
from chainer_addons.training.sacred import SacredTrainerMixin



import re
import logging

from chainer.training import extension

from collections import defaultdict

class AlternateTrainable(extension.Extension):
	def __init__(self, trainable_groups, interval):
		super(AlternateTrainable, self).__init__()
		assert interval >= 1

		self.trainable_groups = trainable_groups
		self.interval = interval
		self.round_interval = interval * len(trainable_groups)

	@property
	def n_groups(self):
		return len(self.trainable_groups)

	def initialize(self, trainer):
		self.model = trainer.updater.get_optimizer("main").target

		group_regexes = [
			(i, re.compile(regex))
				for i, regex in enumerate(self.trainable_groups)
					if regex is not None]

		rest_group_idx = [i
			for i, regex in enumerate(self.trainable_groups)
				if regex is None]

		assert len(rest_group_idx), "There should be only one rest group!"
		rest_group_idx = rest_group_idx[0]

		grouped_params = defaultdict(list)

		for name, param in self.model.namedparams():
			idx = rest_group_idx
			for i, group_regex in group_regexes:
				# if we match a certain name, add corresponding param to the group
				if group_regex.search(name) is not None:
					idx = i
					break
			# else use the "rest" group for the remaining params
			grouped_params[idx].append(param)

		self.grouped_params = dict(grouped_params)

	def finalize(self):
		self.model = None

	def __call__(self, trainer):
		epoch = trainer.updater.epoch
		iteration = trainer.updater.iteration
		is_new = trainer.updater.is_new_epoch or (epoch == 0 and iteration == 1)
		if not is_new: return
		if epoch % self.interval != 0: return

		subinterval = epoch % self.round_interval
		group_idx = subinterval // self.interval

		assert self.model is not None
		logging.debug("Epoch: {epoch}, {subinterval}".format(epoch, subinterval))

		self.enable_group(group_idx)

	def enable_group(self, group_idx):
		logging.debug("Enabling group #{}".format(group_idx))
		self.model.disable_update()
		for param in self.grouped_params[group_idx]:
			rule = param.update_rule
			if rule is not None:
				rule.enabled = True



import numpy as np

from unittest import TestCase, skip

import chainer
import chainer.functions as F

from chainer_addons.links.fisher_encoding import FVLayer
from chainer_addons.functions import fv_encoding, gamma

class BaseTestCase(TestCase):

	def __init__(self, *args, **kwargs):
		super(BaseTestCase, self).__init__(*args, **kwargs)
		self.xp = np

		self.n, self.t, self.in_size = 2, 3, 512
		self.n_components = 3
		self.eps = 0#1e-12

		self.seed = 8934567
		self.dtype = np.float32

	def setUp(self):
		rnd = self.xp.random.RandomState(self.seed)

		X = rnd.randn(self.n, self.t, self.in_size).astype(self.dtype)
		self.X0 = chainer.Variable(X)
		self.X1 = chainer.Variable(X)

		self.init_mu = rnd.randn(self.in_size, self.n_components).astype(self.dtype)
		# self.init_mu = self.xp.ones((self.in_size, self.n_components), self.dtype)

	def _new_fv_layer(self, init_mu=None, init_sig=1):
		return FVLayer(
			self.in_size,
			self.n_components,
			init_mu=self.init_mu if init_mu is None else init_mu,
			init_sig=init_sig,
			eps=self.eps)

class GammaFunctionTest(BaseTestCase):

	def setUp(self):
		super(GammaFunctionTest, self).setUp()

		layer = self._new_fv_layer()

		self.mu = layer.mu
		self.sig = layer.sig
		self.w = layer.w

	def _expand_params(self, params, shape):
		_f = lambda param: \
			F.broadcast_to(
				F.expand_dims(
					F.expand_dims(param, axis=0), axis=0), shape)
		return map(_f, params)

	def _log_proba(self, x, mu, sig):
		shape = (self.n, self.t, self.in_size, self.n_components)

		_x = F.broadcast_to(F.expand_dims(x, -1), shape)
		_mu, _sig = self._expand_params([mu, sig], shape)
		x_mu = F.expand_dims(F.transpose(_x - _mu), axis=1)

		_dist = (_x - _mu) / _sig
		_dist = F.sum(_dist**2, axis=2)
		return -_dist / 2

	def _proba(self, x, mu, sig):
		return F.exp(self._log_proba(x, mu, sig))

	def _reference_gamma(self, x, mu, sig, w):

		_log_proba = self._log_proba(x, mu, sig)
		# self._proba = _proba.array
		_w = F.broadcast_to(F.expand_dims(w, axis=0), _log_proba.shape)
		_log_wu =  _log_proba + F.log(_w)

		_log_wu_sum = F.logsumexp(_log_wu, axis=-1)
		_log_wu_sum = F.expand_dims(_log_wu_sum, -1)
		_log_wu_sum = F.broadcast_to(_log_wu_sum, _log_wu.shape)

		return F.exp(_log_wu - _log_wu_sum)

	def test_gamma_outputs(self):

		gamma0 = self._reference_gamma(self.X0, self.mu, self.sig, self.w)
		gamma1 = gamma(self.X1, self.mu, self.sig, self.w, eps=self.eps)

		self.assertTrue(np.allclose(gamma0.array, gamma1.array),
			"Both implementations should have similar result, but was:\n"+\
			f"{gamma0.array}\n~=\n{gamma1.array}")


		self.assertTrue(np.all(gamma0.array == gamma1.array),
			"Both implementations should have the same result, but was:\n"+\
			f"{gamma0.array}\n!=\n{gamma1.array}")

	@skip
	def test_gamma_gradients(self):

		gamma0 = self._reference_gamma(self.X0, self.mu, self.sig, self.w)
		gamma1 = gamma(self.X1, self.mu, self.sig, self.w, eps=self.eps)

		self.X0.grad, self.X1.grad = None, None

		F.sum(gamma0).backward()
		F.sum(gamma1).backward()

		self.assertIsNotNone(self.X0.grad,
			"Gradient of X0 was not computed!")

		self.assertIsNotNone(self.X1.grad,
			"Gradient of X1 was not computed!")
		# import pdb; pdb.set_trace()

		self.assertTrue(np.allclose(self.X0.grad, self.X1.grad, atol=1e-7),
			"Both computed gradients should be similar, but were:\n" + \
			f"{self.X1.grad}\n~=\n{self.X0.grad}")

		self.assertTrue(np.all(self.X0.grad==self.X1.grad),
			"Both computed gradients should be equal, but were:\n" + \
			f"{self.X0.grad}\n!=\n{self.X1.grad}")

class FVFunctionTests(BaseTestCase):

	def setUp(self):
		super(FVFunctionTests, self).setUp()

		self.layer0 = self._new_fv_layer()
		self.layer1 = self._new_fv_layer()

	def test_layer_mus(self):
		self.assertTrue(np.all(self.layer0.mu == self.layer1.mu),
			"mus should be initialized equally!")

	@skip
	def test_fv_encoding_outputs(self):
		y0 = self.layer0(self.X0)

		mu, sigma, w = self.layer1.mu, self.layer1.sig, self.layer1.w
		y1 = fv_encoding(self.X1, mu, sigma, w, eps=self.eps)

		self.assertTrue(np.allclose(y0.array, y1.array),
			"Both implementations should have similar result, but were:\n"+\
			f"{y0.array}\n~=\n{y1.array}")

		self.assertTrue(np.all(y0.array == y1.array),
			"Both implementations should have the same result, but were:\n"+\
			f"{y0.array}\n!=\n{y1.array}")

	@skip
	def test_fv_encoding_gradients(self):

		y0 = self.layer0(self.X0)

		mu, sigma, w = self.layer1.mu, self.layer1.sig, self.layer1.w
		y1 = fv_encoding(self.X1, mu, sigma, w, eps=self.eps)

		self.X0.grad, self.X1.grad = None, None

		F.sum(y0).backward()
		F.sum(y1).backward()

		self.assertIsNotNone(self.X0.grad,
			"Gradient of X0 was not computed!")

		self.assertIsNotNone(self.X1.grad,
			"Gradient of X1 was not computed!")

		self.assertTrue(np.allclose(self.X0.grad, self.X1.grad, atol=1e-7),
			"Both computed gradients should be similar, but were:\n" + \
			f"{self.X1.grad}\n~=\n{self.X0.grad}")

		self.assertTrue(np.all(self.X0.grad==self.X1.grad),
			"Both computed gradients should be equal, but were:\n" + \
			f"{self.X1.grad}\n!=\n{self.X0.grad}")




import logging
import os
import re
import sys

from sacred import Experiment, SETTINGS
from sacred.observers import MongoObserver
from sacred.utils import apply_backspaces_and_linefeeds

from pathlib import Path
from urllib.parse import quote_plus

def progress_bar_filter(text,
	escape=re.compile(r"\x1B\[([0-?]*[ -/]*[@-~])"),
	line_contents=re.compile(r".*(total|this epoch|Estimated time|\d+ iter, \d+ epoch / \d+ epochs).*\n")):
	_text = apply_backspaces_and_linefeeds(text)
	_text = escape.sub("", _text)
	_text = line_contents.sub("", _text)
	return _text


class MPIExperiment(Experiment):

	def __init__(self, opts, *args,
		host=None,
		comm=None,
		no_observer=False,
		output_filter=progress_bar_filter,
		**kwargs):

		if kwargs.get("base_dir") is None:
			base_dir = Path(sys.argv[0]).resolve().parent
			logging.info(f"Base experiment directory: {base_dir}")
			kwargs["base_dir"] = str(base_dir)

		super(MPIExperiment, self).__init__(*args, **kwargs)

		_main_process = comm is None or comm.rank == 0

		if no_observer or not _main_process:
			return

		self.logger = logging.getLogger()
		self.captured_out_filter = output_filter

		creds = MPIExperiment.get_creds()
		_mongo_observer = MongoObserver.create(
			url=MPIExperiment.auth_url(creds, host=host),
			db_name=creds["db_name"],
		)

		self.observers.append(_mongo_observer)

		self.add_config(**opts.__dict__)

	def __call__(self, *args, **kwargs):
		return self._create_run()(*args, **kwargs)

	@classmethod
	def get_creds(cls):
		return dict(
			user=quote_plus(os.environ["MONGODB_USER_NAME"]),
			password=quote_plus(os.environ["MONGODB_PASSWORD"]),
			db_name=quote_plus(os.environ["MONGODB_DB_NAME"]),
		)

	@classmethod
	def auth_url(cls, creds, host="localhost", port=27017):
		host = host or cls.get_host()
		logging.info(f"MongoDB host: {host}")

		url = "mongodb://{user}:{password}@{host}:{port}/{db_name}?authSource=admin".format(
			host=host, port=port, **creds)
		return url

	@classmethod
	def get_host(cls):
		return quote_plus(os.environ.get("MONGODB_HOST", "localhost"))

SETTINGS.DISCOVER_SOURCES = "dir"

__all__ = [
	"MPIExperiment",
	"progress_bar_filter"
]



from .alpha_prod import alpha_prod
from .signed_square_root import signed_square_root
from .tf_avg_pooling import tf_average_pooling_2d
from .fisher_encoding.fv_encoding import fv_encoding
from .fisher_encoding.gamma import gamma

import chainer.functions as F

def smoothed_cross_entropy(pred, gt, N, eps=0.1):

	loss = F.softmax_cross_entropy(pred, gt)
	# -sum[ log( P(k) ) * U ]
	reg_loss = -F.mean(F.sum(F.log_softmax(pred) / N, axis=1))

	return (1-eps) * loss + eps * reg_loss




import numpy as np
from chainer.backends import cuda

from .base import GPUMixin

from sklearn.mixture.bayesian_mixture import BayesianGaussianMixture

class BaysianGMM(GPUMixin, BayesianGaussianMixture):

	def _m_step(self, X, log_resp, xp=np):

		nk, means, covariances = \
			self._gaussian_params(X, log_resp, xp=xp)

		# estimate weights
		self.weight_concentration_ = self.weight_concentration_prior_ + nk
		self.weights_ = self.weight_concentration_ / self.weight_concentration_.sum()

		# estimate means
		self.mean_precision_ = self.mean_precision_prior_ + nk
		self.means_ = ((self.mean_precision_prior_ * self.mean_prior_ +
						nk[:, None] * means) /
					   self.mean_precision_[:, None])

		# estimate precisions
		_, n_features = means.shape

		# Warning : in some Bishop book, there is a typo on the formula 10.63
		# `degrees_of_freedom_k = degrees_of_freedom_0 + Nk`
		# is the correct formula
		self.degrees_of_freedom_ = self.degrees_of_freedom_prior_ + nk

		diff = means - self.mean_prior_
		self.covariances_ = (
			self.covariance_prior_ + nk[:, None] * (
				covariances + (self.mean_precision_prior_ /
					  self.mean_precision_)[:, None] * xp.square(diff)))

		# Contrary to the original bishop book, we normalize the covariances
		self.covariances_ /= self.degrees_of_freedom_[:, None]
		self.precisions_cholesky_ = 1. / xp.sqrt(self.covariances_)

	def _check_parameters(self, X):
		super(BaysianGMM, self)._check_parameters(cuda.to_cpu(X))
		xp = self.xp_from_array(X)

		self.mean_prior_ = xp.array(self.mean_prior_)
		self.covariance_prior_ = xp.array(self.covariance_prior_)


	# def fit(self, X, y=None):
	# 	X, xp = self._tranform_X(X)
	# 	return super(BayesianGaussianMixture, self).fit(X, y)




from chainer_addons.training.extensions import SacredReport

class SacredTrainerMixin(object):
	def __init__(self, opts, sacred_params, sacred_trigger=(1, "epoch"), *args, **kwargs):
		super(SacredTrainerMixin, self).__init__(opts=opts, *args, **kwargs)
		self.sacred_reporter = SacredReport(opts,
			sacred_params=sacred_params,
			trigger=sacred_trigger)
		self.extend(self.sacred_reporter)

		def _run(*args, **kwargs):
			res = super(SacredTrainerMixin, self).run(*args, **kwargs)

			# import pdb; pdb.set_trace()
			return res

		if getattr(self, "_only_eval", False):
			self.run = _run
		else:
			self.sacred_reporter.ex.main(_run)
			self.run = self.sacred_reporter.ex




from chainer.function import Function
from chainer import cuda

from .base import GammaMixin


class FVEncoding(Function, GammaMixin):

	def forward(self, inputs):
		xp = cuda.get_array_module(*inputs)

		x, mu, sigma, w = inputs
		n, t, in_size = x.shape
		_, n_components = mu.shape

		_gamma = self._gamma(inputs)

		self.gamma = xp.broadcast_to(
			xp.expand_dims(_gamma, axis=2), self._x_mu.shape)

		_x_mu_sig = self._x_mu / self._sig

		G_mu = xp.sum(self.gamma * _x_mu_sig, axis=1)
		G_sig = xp.sum(self.gamma * (_x_mu_sig**2 - 1), axis=1)

		_w = xp.broadcast_to(w, (n, in_size, n_components))

		G_mu /= n_components * xp.sqrt(_w)
		G_sig /= n_components * xp.sqrt(2 * _w)

		# 2 * (n, in_size, n_components) -> (n, in_size, n_components, 2) -> (n, 2*in_size*n_components)
		res = xp.stack([G_mu, G_sig], axis=-1).reshape(n, -1)
		# res = xp.stack([G_mu], axis=-1).reshape(n, -1)
		return res,


	def backward(self, inputs, gys):
		xp = cuda.get_array_module(*inputs)
		x, mu, sigma, w = inputs

		n, t, in_size = x.shape
		_, n_components = mu.shape

		# ##### gamma-Gradient computation #####
		g_gamma = self._gamma_grad(inputs)

		##### G_mu- and G_sig-Gradient computation #####
		_x_mu, _sig = self._x_mu, self._sig

		shape = (n, t, in_size, n_components)
		_w = xp.broadcast_to(w[None][None][None], shape)
		g_G_mu  = self.gamma * (g_gamma * _x_mu + 1) / (n_components * xp.sqrt(_w) * _sig)

		g_G_sig = self.gamma * (
				g_gamma * (_x_mu**2 / _sig**2 - 1) + \
				2 * _x_mu / _sig**2
			) / (n_components * xp.sqrt(_w * 2))

		##### reshaping according to the output gradients #####
		gx = xp.stack([g_G_mu, g_G_sig], axis=-1)
		# gx = xp.stack([g_G_mu], axis=-1)
		_gy = xp.reshape(gys[0], (n, in_size, n_components, gx.shape[-1]))
		_gy = xp.expand_dims(_gy, axis=1)
		gx = xp.sum(gx * _gy, axis=(-1, -2))

		grads = gx, None, None, None
		return grads

def fv_encoding(x, mu, sigma, w, eps=None):
	return FVEncoding(eps=eps)(x, mu, sigma, w)



from chainer.training import Extension
from chainer.training import trigger as trigger_module


class WarmUp(Extension):

	def __init__(self, epochs, model, initial_lr, warm_up_lr=1.0):
		super(WarmUp, self).__init__()
		self.epochs = epochs
		self._trigger = trigger_module.get_trigger((epochs, "epoch"))

		self.lr = initial_lr
		self.warm_up_lr = warm_up_lr
		self.model = model
		assert hasattr(self.model, "clf_layer"), \
			"Model instance should have \"clf_layer\" attribute!"

		self._done = False

	def initialize(self, trainer):
		self.model.disable_update()
		self.model.clf_layer.enable_update()

		opt = trainer.updater.get_optimizer("main")
		if hasattr(opt, "alpha"):
			opt.alpha = self.warm_up_lr
		else:
			opt.lr = self.warm_up_lr

	def __call__(self, trainer):
		if not self._trigger(trainer): return
		if self._done: return

		opt = trainer.updater.get_optimizer("main")
		if hasattr(opt, "alpha"):
			opt.alpha = self.lr
		else:
			opt.lr = self.lr

		self.model.enable_update()
		self._done = True




def all_children(model):
	"""
		Iterate over all links in a (nested) chain.
	"""

	for child in model.children():
		grand_children = list(child.children())
		if grand_children:
			for grand_child in all_children(child):
				yield grand_child
		else:
			yield child

import os
import logging, simplejson as json
from os.path import isdir, isfile, join

def prepare_output(output, args):

	if not isdir(output):
		os.makedirs(output)

	if not isfile(join(output, ".nobackup")):
		open(join(output, ".nobackup"), "w").close()

	args_dump = join(output, "args.json")
	logging.info("Saving arguments to \"{}\"".format(args_dump))
	with open(args_dump, "w") as f:
		json.dump(args, f, indent=2)


from .imgproc import Size



from chainer.optimizer_hooks import WeightDecay

class SelectiveWeightDecay(WeightDecay):

	def __init__(self, rate, selection=None):
		self.rate = rate
		self.selection = selection or (lambda _: True)

	def __call__(self, rule, param):
		if not self.selection(param): return
		return super(SelectiveWeightDecay, self).__call__(rule, param)





import numpy as np
from chainer_addons.training.extensions.learning_rate.base import ContiniousLearningRate

class CosineAnnealingLearningRate(ContiniousLearningRate):

	def __init__(self, stages=4, *args, **kwargs):
		super(CosineAnnealingLearningRate, self).__init__(*args, **kwargs)
		self.stages = stages
		self._init_epochs = self._epochs
		self._epochs = self._init_epochs * (2**stages - 1)

		self.current_stage = 0

	@property
	def epochs_in_current_stage(self):
		return self._init_epochs * 2**self.current_stage

	@property
	def end_of_current_stage(self):
		return self._init_epochs * (2**(self.current_stage + 1) - 1)

	@property
	def current_stage_offset(self):
		return self._init_epochs * (2**self.current_stage - 1)

	@property
	def factor(self):
		return 0.5 * (self._lr - self._target)

	def new_lr(self, epoch):

		if epoch >= self.end_of_current_stage:
			self.current_stage += 1

		epoch -= self.current_stage_offset
		epochs = self.epochs_in_current_stage - self._offset

		res = self._target + self.factor * (1 + np.cos(np.pi * epoch / epochs))

		return res


if __name__ == '__main__':

	from matplotlib import pyplot as plt
	ext = CosineAnnealingLearningRate(
		attr="attr",
		lr=1e-2,
		target=1e-5,
		epochs=12,
		offset=0,
		stages=5)

	xs = np.arange(ext._epochs).astype(np.float32)
	ys = np.zeros_like(xs)

	for x in xs:
		ys[int(x)] = ext.new_lr(x)

	plt.plot(xs, ys)
	plt.show()



import chainer
import chainer.functions as F
import chainer.links as L

from chainer_addons.links import Conv2D_BN
from chainer_addons.links.pooling import PoolingType


class InceptionHead(chainer.Chain):
	def __init__(self):
		super(InceptionHead, self).__init__()
		with self.init_scope():
			# input 3 x 299 x 299
			self.conv1 = Conv2D_BN(3,   32, ksize=3, stride=2)
			# input 32 x 149 x 149
			self.conv2 = Conv2D_BN(32,  32, ksize=3)
			# input 32 x 147 x 147
			self.conv3 = Conv2D_BN(32,  64, ksize=3, pad=1)
			# input 64 x 147 x 147
			self.pool4 = PoolingType.new("max", ksize=3, stride=2)

			# input 64 x 73 x 73
			self.conv5 = Conv2D_BN(64,  80, ksize=1)
			# input 80 x 73 x 73
			self.conv6 = Conv2D_BN(80, 192, ksize=3)
			# input 192 x 71 x 71
			self.pool7 = PoolingType.new("max", ksize=3, stride=2)
			# output 192 x 35 x 35

	def __call__(self, x):
		x = self.conv1(x)
		x = self.conv2(x)
		x = self.conv3(x)
		x = self.pool4(x)
		x = self.conv5(x)
		x = self.conv6(x)
		x = self.pool7(x)
		return x

class Inception1(chainer.Chain):
	def __init__(self, insize, sizes, outputs, **pool_args):
		super(Inception1, self).__init__()

		out1x1, out5x5, out3x3, out_pool = outputs
		s5x5, s3x3_1, s3x3_2 = sizes

		with self.init_scope():
			self.conv1x1   = Conv2D_BN(insize, out1x1, ksize=1)

			self.conv5x5_1 = Conv2D_BN(insize, s5x5,   ksize=1)
			self.conv5x5_2 = Conv2D_BN(s5x5,   out5x5, ksize=5, pad=2)

			self.conv3x3_1 = Conv2D_BN(insize, s3x3_1, ksize=1)
			self.conv3x3_2 = Conv2D_BN(s3x3_1, s3x3_2, ksize=3, pad=1)
			self.conv3x3_3 = Conv2D_BN(s3x3_2, out3x3, ksize=3, pad=1)

			self.pool_conv = Conv2D_BN(insize, out_pool, ksize=1)
		self.pool = PoolingType.new(**pool_args)

	def __call__(self, x):

		y0 = self.conv1x1(x)
		y1 = self.conv5x5_2(self.conv5x5_1(x))
		y2 = self.conv3x3_3(self.conv3x3_2(self.conv3x3_1(x)))
		y3 = self.pool_conv(self.pool(x))
		return F.concat([y0, y1, y2, y3])

class Inception2(chainer.Chain):
	def __init__(self, insize, sizes, outputs, **pool_args):
		super(Inception2, self).__init__()

		out1, out2 = outputs
		size1, size2 = sizes
		with self.init_scope():
			self.conv3x3   = Conv2D_BN(insize, out1 , ksize=3, stride=2)

			self.conv3x3_1 = Conv2D_BN(insize, size1, ksize=1)
			self.conv3x3_2 = Conv2D_BN(size1,  size2, ksize=3, pad=1)
			self.conv3x3_3 = Conv2D_BN(size2,  out2,  ksize=3, stride=2)

		self.pool = PoolingType.new(**pool_args)

	def __call__(self, x):
		y0 = self.conv3x3(x)
		y1 = self.conv3x3_3(self.conv3x3_2(self.conv3x3_1(x)))
		y2 = self.pool(x)
		return F.concat([y0, y1, y2])

class Inception3(chainer.Chain):
	def __init__(self, insize, sizes, outputs, **pool_args):
		super(Inception3, self).__init__()

		out1x1, out7x7, out7x7x2, out_pool = outputs
		s7x7_1, s7x7_2, s7x7x2_1, s7x7x2_2, s7x7x2_3, s7x7x2_4 = sizes

		with self.init_scope():
			self.conv1x1 = Conv2D_BN(insize,       out1x1, ksize=1)

			self.conv7x7_1 = Conv2D_BN(insize,     s7x7_1, ksize=1)
			self.conv7x7_2 = Conv2D_BN(s7x7_1,     s7x7_2, ksize=(1,7), pad=(0,3))
			self.conv7x7_3 = Conv2D_BN(s7x7_2,     out7x7, ksize=(7,1), pad=(3,0))

			self.conv7x7x2_1 = Conv2D_BN(insize,   s7x7x2_1, ksize=1)
			self.conv7x7x2_2 = Conv2D_BN(s7x7x2_1, s7x7x2_2, ksize=(7,1), pad=(3,0))
			self.conv7x7x2_3 = Conv2D_BN(s7x7x2_2, s7x7x2_3, ksize=(1,7), pad=(0,3))
			self.conv7x7x2_4 = Conv2D_BN(s7x7x2_3, s7x7x2_4, ksize=(7,1), pad=(3,0))
			self.conv7x7x2_5 = Conv2D_BN(s7x7x2_4, out7x7x2, ksize=(1,7), pad=(0,3))

			self.pool_conv = Conv2D_BN(insize, out_pool, ksize=1)

		self.pool = PoolingType.new(**pool_args)

	def __call__(self, x):
		y0 = self.conv1x1(x)
		y1 = self.conv7x7_3(self.conv7x7_2(self.conv7x7_1(x)))
		y2 = self.conv7x7x2_5(self.conv7x7x2_4(self.conv7x7x2_3(self.conv7x7x2_2(self.conv7x7x2_1(x)))))
		y3 = self.pool_conv(self.pool(x))

		return F.concat([y0, y1, y2, y3])

class Inception4(chainer.Chain):
	def __init__(self, insize, sizes, outputs, **pool_args):
		super(Inception4, self).__init__()

		out3x3, out7x7 = outputs
		s3x3, s7x7_1, s7x7_2, s7x7_3 = sizes

		with self.init_scope():
			self.conv3x3_1 = Conv2D_BN(insize, s3x3, ksize=1)
			self.conv3x3_2 = Conv2D_BN(s3x3, out3x3, ksize=3, stride=2)

			self.conv7x7_1 = Conv2D_BN(insize, s7x7_1, ksize=1)
			self.conv7x7_2 = Conv2D_BN(s7x7_1, s7x7_2, ksize=(1, 7), pad=(0, 3))
			self.conv7x7_3 = Conv2D_BN(s7x7_2, s7x7_3, ksize=(7, 1), pad=(3, 0))
			self.conv7x7_4 = Conv2D_BN(s7x7_3, out7x7, ksize=3, stride=2)

		self.pool = PoolingType.new(**pool_args)

	def __call__(self, x):

		y0 = self.conv3x3_2(self.conv3x3_1(x))
		y1 = self.conv7x7_4(self.conv7x7_3(self.conv7x7_2(self.conv7x7_1(x))))
		y2 = self.pool(x)

		return F.concat([y0, y1, y2])

class Inception5(chainer.Chain):
	def __init__(self, insize, sizes, outputs, **pool_args):
		super(Inception5, self).__init__()

		out1x1, out3x3, out3x3x2, out_pool = outputs
		s3x3, s3x3x2_1, s3x3x2_2 = sizes

		with self.init_scope():

			self.conv1x1 = Conv2D_BN(insize, out1x1, ksize=1)

			self.conv3x3_1 = Conv2D_BN(insize, s3x3, ksize=1)
			self.conv3x3_2 = Conv2D_BN(s3x3, out3x3, ksize=(1, 3), pad=(0,1))
			self.conv3x3_3 = Conv2D_BN(s3x3, out3x3, ksize=(3, 1), pad=(1,0))

			self.conv3x3x2_1 = Conv2D_BN(insize  , s3x3x2_1, ksize=1)
			self.conv3x3x2_2 = Conv2D_BN(s3x3x2_1, s3x3x2_2, ksize=3, pad=1)
			self.conv3x3x2_3 = Conv2D_BN(s3x3x2_2, out3x3x2, ksize=(1, 3), pad=(0,1))
			self.conv3x3x2_4 = Conv2D_BN(s3x3x2_2, out3x3x2, ksize=(3, 1), pad=(1,0))

			self.pool_conv = Conv2D_BN(insize, out_pool, ksize=1)

		self.pool = PoolingType.new(**pool_args)


	def __call__(self, x):
		y0 = self.conv1x1(x)


		y1 = self.conv3x3_1(x)
		y1 = F.concat([self.conv3x3_2(y1), self.conv3x3_3(y1)])

		y2 = self.conv3x3x2_2(self.conv3x3x2_1(x))
		y2 = F.concat([self.conv3x3x2_3(y2), self.conv3x3x2_4(y2)])

		y3 = self.pool_conv(self.pool(x))

		return F.concat([y0, y1, y2, y3])


class AuxilaryClassifier(chainer.Chain):
	def __init__(self, n_classes):
		super(AuxilaryClassifier, self).__init__()
		with self.init_scope():
			self.conv1 = Conv2D_BN(768, 128, ksize=1, pad=1)
			self.conv2 = Conv2D_BN(128, 768, ksize=7)

			self.fc = L.Linear(n_classes)


	def __call__(self, x):
		x = F.average_pooling_2d(x, ksize=5, stride=3)
		x = self.conv1(x)
		x = self.conv2(x)
		return self.fc(x)




import chainer
from chainer.backends import cuda
from chainer import link
from chainer import initializers
from chainer import functions as F

from chainer_addons.links.fisher_encoding.mixtures import GMM

import numpy as np

_LOG_2PI = np.log(2 * np.pi)

class GMMLayer(link.Link):

	def __init__(self, in_size, n_components,
		init_mu=None, init_sig=1,
		init_from_data=False,
		alpha=0.9, eps=1e-2,
		gmm_cls=GMM,
		use_sk_learn_em=False,
		sk_learn_kwargs={},
		dtype=np.float32):

		super(GMMLayer, self).__init__()

		self.n_components = n_components
		self.in_size = in_size
		self.use_sk_learn_em = use_sk_learn_em

		self.i = 0
		self.lim = 1
		self.visualization_interval = 100
		self.visualization_folder = "mu_change"

		if init_mu is None:
			self.init_mu = initializers.Uniform(scale=1, dtype=dtype)

		elif isinstance(init_mu, (int, float)):
			self.init_mu = initializers.Uniform(scale=init_mu, dtype=dtype)

		elif isinstance(init_mu, (np.ndarray, cuda.ndarray)):
			self.init_mu = initializers.Constant(init_mu, dtype=dtype)

		elif not isinstance(init_mu, chainer.initializer.Initializer):
			raise ValueError(
				"\"init_mu\" should be either omited, be an instance of " + \
				"chainer.initializer.Initializer or a numpy/cupy array!"
			)


		self.init_sig = initializers.Constant(init_sig, dtype=dtype)
		self.init_w = initializers.Constant(1 / n_components, dtype=dtype)

		with self.init_scope():
			self.add_persistent("mu",
				np.zeros((self.in_size, self.n_components), dtype))

			self.add_persistent("sig",
				np.zeros((self.in_size, self.n_components), dtype))

			self.add_persistent("w",
				np.zeros((self.n_components), dtype))

			self.add_persistent("alpha", alpha)
			self.add_persistent("eps", eps)
			self.add_persistent("t", 1)

			# for key, value in sk_learn_kwargs.items():
			# 	self.add_persistent(key, value)

			self.init_params()

		self._initialized = not init_from_data

		self.gmm_cls = gmm_cls
		if "reg_covar" not in sk_learn_kwargs:
			sk_learn_kwargs["reg_covar"] = eps
		self.sk_learn_kwargs = sk_learn_kwargs
		self.sk_gmm = None

	@property
	def precisions_chol(self):
		"""
			Compute the Cholesky decomposition of the precisions.
			Reference:
				https://github.com/scikit-learn/scikit-learn/blob/0.21.3/sklearn/mixture/gaussian_mixture.py#L288
		"""
		return 1. / self.xp.sqrt(self.sig)

	def reset(self):
		self.t = 1

	def init_params(self):
		self.init_mu(self.mu)
		self.init_sig(self.sig)
		self.init_w(self.w)

	def init_from_data(self, x, gmm_cls=None):

		if isinstance(x, chainer.Variable):
			data = cuda.to_cpu(x.array)
		else:
			data = cuda.to_cpu(x)

		gmm = self.new_gmm(gmm_cls=gmm_cls)
		self.set_gmm_params(gmm)

		gmm.fit(data.reshape(-1, data.shape[-1]))

		self.mu[:]  = self.xp.array(gmm.means_.T)
		self.sig[:] = self.xp.array(gmm.covariances_.T)
		self.w[:]   = self.xp.array(gmm.weights_)

		self._initialized = True

	def new_gmm(self, gmm_cls=None, *args, **kwargs):
		return (gmm_cls or self.gmm_cls)(
			covariance_type="diag",
			n_components=self.n_components,
			**kwargs
		)

	def set_gmm_params(self, gmm):
		means_, covariances_, prec_chol_, weights_ = \
			[self.mu.T, self.sig.T, self.precisions_chol.T, self.w]
			# map(cuda.to_cpu, [self.mu.T, self.sig.T, self.precisions_chol.T, self.w])

		gmm.precisions_cholesky_ = prec_chol_
		gmm.covariances_ = covariances_
		gmm.means_= means_
		gmm.weights_= weights_

	def as_sklearn_gmm(self, gmm_cls=None, **gmm_kwargs):
		gmm = self.new_gmm(gmm_cls=gmm_cls, warm_start=True, **gmm_kwargs)
		self.set_gmm_params(gmm)
		return gmm


	def _check_input(self, x):
		assert x.ndim == 3, \
			"input should have following dimensions: (batch_size, n_features, feature_size)"
		n, t, in_size = x.shape
		assert in_size == self.in_size, \
			"feature size of the input does not match input size: ({} != {})! ".format(
				in_size, self.in_size)
		return n, t


	def _comp_params(self, i):
		return self.mu[:, i], self.sig[:, i], self.w[i]

	def _expand_params(self, x):
		n, t = self._check_input(x)
		shape = (n, t, self.in_size, self.n_components)
		shape2 = (n, t, self.n_components)

		_x = F.broadcast_to(F.expand_dims(x, -1), shape)

		_params = [(self.mu, shape), (self.sig, shape), (self.w, shape2)]
		_mu, _sig, _w = [
			F.broadcast_to(F.expand_dims(F.expand_dims(p, 0), 0), s)
				for p, s in _params]

		return _x, _mu, _sig, _w

	def soft_assignment(self, x):
		# _log_proba, _w = self.log_proba(x, weighted=False)
		# _log_wu = _log_proba + F.log(_w)

		# _proba = F.exp(_log_wu - _log_wu.array.max(axis=-1))
		# _proba /= _proba.array.sum(axis=-1)
		# return _proba

		return F.exp(self.log_soft_assignment(x))

	def log_soft_assignment(self, x):

		_log_proba, _w = self.log_proba(x, weighted=False)
		_log_wu = _log_proba + F.log(_w)

		_log_wu_sum = F.logsumexp(_log_wu, axis=-1)
		_log_wu_sum = F.expand_dims(_log_wu_sum, axis=-1)
		_log_wu_sum = F.broadcast_to(_log_wu_sum, _log_wu.shape)

		return _log_wu - _log_wu_sum

	def _sk_learn_dist(self, x):
		"""
			Estimate the log Gaussian probability
			Reference:
				https://github.com/scikit-learn/scikit-learn/blob/0.21.3/sklearn/mixture/gaussian_mixture.py#L380
		"""
		n, t, size = x.shape
		_x = x.reshape(-1, size).array
		_mu = self.mu.T
		_precs = 1 / self.sig.T

		res0 = F.sum((_mu ** 2 * _precs), 1)
		res1 = -2. * F.matmul(_x, (_mu * _precs).T)
		res2 = F.matmul(_x ** 2, _precs.T)
		res = F.broadcast_to(res0, res1.shape) + res1 + res2
		return res.reshape(n, t, -1)


	def log_proba(self, x, weighted=False, use_sk_learn=False):

		_x, _mu, _sig, _w = self._expand_params(x)

		if use_sk_learn:
			_dist = self._sk_learn_dist(x)
			prec_chol_ = self.precisions_chol
			# det(precision_chol) is half of det(precision)
			log_det_chol = self.xp.sum(self.xp.log(prec_chol_), axis=0)
			_log_proba = -0.5 * (self.in_size * _LOG_2PI + _dist) + log_det_chol
		else:
			_dist = F.sum((_x - _mu)**2 / _sig, axis=2)
			# normalize with (2*pi)^k and det(sig) = prod(diagonal_sig)
			log_det = self.xp.sum(self.xp.log(self.sig), axis=0)
			_log_proba = -0.5 * (self.in_size * _LOG_2PI + _dist + log_det)

		# mask log probs greater than 0!
		# _log_proba *= _log_proba.array <= 0

		# if self.xp.any(_log_proba.array > 0):
		# 	import pdb; pdb.set_trace()


		if weighted:
			_log_wu = _log_proba + F.log(_w)
			_log_proba = F.logsumexp(_log_wu, axis=-1)

		return _log_proba, _w


	def proba(self, *args, **kwargs):
		_log_proba, _w = self.log_proba(*args, **kwargs)
		return F.exp(_log_proba), _w

	def __call__(self, x, use_mask=False, visibility_mask=None):
		if chainer.config.train:
			mask = self.get_mask(x,
			                     use_mask=use_mask,
			                     visibility_mask=visibility_mask)
			selected = x[mask]
			selected = selected.reshape(-1, x.shape[-1])
			self.update_parameter(selected)
		return x

	def get_mask(self, x, use_mask, visibility_mask=None):
		if not use_mask: return Ellipsis
		_feats = x.array if hasattr(x, "array") else x
		_feat_lens = self.xp.sqrt(self.xp.sum(_feats**2, axis=2))

		if visibility_mask is None:
			_mean_feat_lens = _feat_lens.mean(axis=1, keepdims=True)
			selected = _feat_lens >= _mean_feat_lens
		else:
			_mean_feat_lens = _feat_lens.sum(axis=1, keepdims=True)
			_n_visible_feats = visibility_mask.sum(axis=1, keepdims=True).astype(np.float32)
			_mean_feat_lens /= _n_visible_feats
			selected = _feat_lens >= _mean_feat_lens
			selected = self.xp.logical_and(selected, visibility_mask)

		if selected.sum() == 0:
			raise RuntimeError("Selection was empty!")

		return self.xp.where(selected)

	def _ema(self, old, new):
		prev_correction = 1 - (self.alpha ** (self.t-1))
		correction = 1 - (self.alpha ** self.t)

		uncorrected_old = old * prev_correction
		res = self.alpha * uncorrected_old + (1 - self.alpha) * new

		return res / correction

	def get_new_params(self, x):
		if self.use_sk_learn_em:
			if self.sk_gmm is None:
				# self.sk_gmm = self.as_sklearn_gmm(**self.sk_learn_kwargs)
				self.sk_gmm = self.new_gmm(**self.sk_learn_kwargs)

			self.sk_gmm.fit(x)

			new_mu, new_sig, new_w = map(self.xp.array,
				[self.sk_gmm.means_.T, self.sk_gmm.covariances_.T, self.sk_gmm.weights_.T])

		else:
			raise NotImplementedError("This code has some issues! Use scikit-learn EM option!")
			## E-Step ##
			gamma = self.soft_assignment(x)
			n, t, n_components = gamma.shape

			## M-Step ##
			N_K = gamma.array.sum(axis=(0, 1))
			N_K = self.xp.maximum(N_K, self.eps)

			# compute new weights
			new_w =  N_K / N_K.sum()

			# compute new mus
			shape = (n, t, self.in_size, self.n_components)
			_x = F.broadcast_to(F.expand_dims(x, axis=3), shape)
			_gamma = F.broadcast_to(F.expand_dims(gamma, axis=2), shape)

			new_mu = (_x * _gamma).array.sum(axis=(0, 1)) / N_K

			# compute new sigmas
			###### estimating sigma does not work yet! ######
			_dist = (_x.array - new_mu)**2
			new_sig = (_dist * _gamma.array).sum(axis=(0, 1)) / N_K

		return new_mu, new_sig, new_w

	def update_parameter(self, x):
		if not self._initialized:
			self.init_from_data(x)

		if self.alpha >= 1:
			return

		new_mu, new_sig, new_w = self.get_new_params(x)

		self.t += 1
		self.w[:] = self._ema(self.w, new_w)
		self.mu[:]  = self._ema(self.mu, new_mu)
		self.sig[:] = self._ema(self.sig, new_sig)

		self.sig = self.xp.maximum(self.sig, self.eps)

		# self.i += 1
		# if (self.i-1) % self.visualization_interval == 0:
		# 	self.__visualize(x, gamma, new_mu, None, new_w)

	def __visualize(self, x, soft_assignment, *new_params):

		import os
		from os.path import join, isdir
		import matplotlib.pyplot as plt
		mu, sig, w = new_params

		fig, ax = plt.subplots()

		data = cuda.to_cpu(x.array).reshape(-1, x.shape[-1])
		data_col = cuda.to_cpu(soft_assignment.array).argmax(axis=-1).reshape(-1)
		xs, ys = data.T
		cmap = plt.cm.jet

		ax.scatter(xs, ys, alpha=0.6, c=data_col, cmap=cmap)
		_new_mus, _old_mus = cuda.to_cpu(mu), cuda.to_cpu(self.mu)

		# ax.scatter(*_new_mus, marker="x", c="black")
		# ax.scatter(*_old_mus, marker="o", c="red")

		for i, ((x0, y0), (x1, y1)) in enumerate(zip(_new_mus.T, _old_mus.T)):
			col = np.array(cmap(i / self.n_components))

			ax.scatter(x0, y0, marker="x", c=col.reshape(1,-1))
			ax.scatter(x1, y1, marker="D", c=col.reshape(1,-1))

			ax.plot([x0, x1], [y0, y1], c=col)


		self.lim = max(self.lim, float(self.xp.abs(x.array).max())*1.05)
		ax.set_xlim(-self.lim, self.lim)
		ax.set_ylim(-self.lim, self.lim)

		if not isdir(self.visualization_folder):
			os.makedirs(self.visualization_folder)

		ax.set_title("Iteration {:06d}".format(self.i))
		plt.tight_layout()
		fig.savefig(join(self.visualization_folder, "iteration_{:06d}.png".format(self.i)))

		# plt.show()
		plt.close()
		# import pdb; pdb.set_trace()


